vg.Model = (function() {
  function model() {
    this._defs = null;
    this._data = {};
    this._scene = null;
    this._reset = {axes: false, legends: false};
  }

  var prototype = model.prototype;

  prototype.defs = function(defs) {
    if (!arguments.length) return this._defs;
    this._defs = defs;
    return this;
  };

  prototype.data = function(data) {
    if (!arguments.length) return this._data;

    var deps = {},
        defs = this._defs,
        src  = defs.data.source,
        tx   = defs.data.flow || {},
        keys = defs.data.sorted,
        len  = keys.length, i, k, x;

    // collect source data set dependencies
    function sources(k) {
      (src[k] || []).forEach(function(s) { deps[s] = k; sources(s); });
    }
    vg.keys(data).forEach(sources);
    
    // update data sets in dependency-aware order
    for (i=0; i<len; ++i) {
      if (data[k=keys[i]]) {
        x = data[k];
      } else if (deps[k]) {
        x = vg_data_duplicate(this._data[deps[k]]);
        if (vg.isTree(data)) vg_make_tree(x);
      } else continue;
      this._data[k] = tx[k] ? tx[k](x, this._data, defs.marks) : x;
    }

    this._reset.legends = true;
    return this;
  };

  prototype.width = function(width) {
    if (this._defs) this._defs.width = width;
    if (this._defs && this._defs.marks) this._defs.marks.width = width;
    if (this._scene) this._scene.items[0].width = width;
    this._reset.axes = true;
    return this;
  };

  prototype.height = function(height) {
    if (this._defs) this._defs.height = height;
    if (this._defs && this._defs.marks) this._defs.marks.height = height;
    if (this._scene) this._scene.items[0].height = height;
    this._reset.axes = true;
    return this;
  };

  prototype.scene = function(node) {
    if (!arguments.length) return this._scene;
    this._scene = node;
    return this;
  };

  prototype.build = function() {
    var m = this, data = m._data, marks = m._defs.marks;
    m._scene = vg.scene.build.call(m, marks, data, m._scene);
    m._scene.items[0].width = marks.width;
    m._scene.items[0].height = marks.height;
    m._scene.interactive = false;
    return this;
  };

  prototype.encode = function(trans, request, item) {
    this.reset();
    var m = this, scene = m._scene, defs = m._defs;
    vg.scene.encode.call(m, scene, defs.marks, trans, request, item);
    return this;
  };

  prototype.reset = function() {
    if (this._scene && this._reset.axes) {
      vg.scene.visit(this._scene, function(item) {
        if (item.axes) item.axes.forEach(function(axis) { axis.reset(); });
      });
      this._reset.axes = false;
    }
    if (this._scene && this._reset.legends) {
      vg.scene.visit(this._scene, function(item) {
        if (item.legends) item.legends.forEach(function(l) { l.reset(); });
      });
      this._reset.legends = false;
    }
    return this;
  };

  return model;
})();