vg.data.impute = function() {
  var domain  = null,
      groupby = null,
      sortby  = null,
      field   = "data",
      method  = "value", // value | mean | prev | next | linear
      value   = 0,
      missing = null,
      gaccess = null,
      faccess = null;

  function get_value() { return value; }

  var imputers = {
    value: get_value,
    mean:  get_value,
    prev: function(idx, prev, next, pidx, nidx) {
    	  return prev;
      },
    next: function(idx, prev, next, pidx, nidx) {
    	  return next;
      },
    linear: function(idx, prev, next, pidx, nidx) {
    	  return prev + (next - prev) * (idx-pidx) / (nidx - pidx);
      }
  };

  function impute(data) {
    // partition and sort data, impute missing tuples
    var groups = group(data);
    
    // impute missing values per group
    for (var i=0; i<groups.length; ++i) {
      process(groups[i]);
    }
    
    // merge groups and return
    return Array.concatenate.apply([], groups);
  }
  
  function process(data) {
    var old_value = value,
        imp = imputers[method],
        len = data.length;
        i, start, end, prev, next, curr;

    if (method === 'mean') {
      value = d3.mean(data, function(x) {
        var val = faccess(x);
        return val == missing ? null : val;
      });
    }

    for (start=0; start<len; ++start) {
      // determine contiguous range of missing values
      for (end=i, end<len && faccess(data[end])==missing, ++end) {}
      prev = data[start-1];
      next = data[end];
      
      // impute values within range
      for (curr=start; curr<end; ++curr) {
        imp(curr, prev, next, start-1, end, data);
      }
      
      start = end;
    }
    
    value = old_value;
  }
  
  function group(data) {
    var groups = [],
        map, i, x, k, g, s, max;

    // TODO: consult domain, impute missing tuples as needed
    if (domain != null) {
      // loop over data, collect all domain values
      // sort domain values
      // create value-index lookup map
    }

    if (groupby == null) {
      groups.push(data.slice());
    } else {
      for (map={}, i=0; i<data.length; ++i) {
        x = data[i];
        // TODO if domain, collect key values as well
        // create pre-sized array, slot value into proper location
        // TODO figure out relationship between domain and sort...
        // (add missing_records flag?)
        k = vg.keystr(gaccess.map(function(f) { return f(x); }));
        g = map[k] || (groups.push(map[k] = []), map[k]);
        g.push(x);
      }
    }

    if (sortby != null) {
      for (i=0; i<groups.length; ++i) {
        groups[i].sort(sortby);
      }
    }

    return groups;
  }

  impute.groupby = function(f) {
    groupby = vg.array(f);
    gaccess = groupby.map(vg.accessor(x));
    return impute;
  };

  impute.sortby = function(s) {
    sortby = vg.comparator(s);
    return impute;
  };
  
  // TODO add missing parameters

  return impute;
};